@@ -23,6 +23,7 @@
 
 #include "ephy-debug.h"
 #include "ephy-file-helpers.h"
+#include "ephy-profile-utils.h"
 #include "ephy-settings.h"
 
 #include <errno.h>
@@ -31,6 +32,7 @@
 #include <libsoup/soup.h>
 #include <stdlib.h>
 #include <string.h>
+#include <fcntl.h>
 
 /* Web Apps are installed in the default config dir of the user.
  * Every app has its own profile directory. To create a web app
@@ -39,7 +41,7 @@
  *
  * The id is used to uniquely identify the app.
  *  - Program name: epiphany-<id>
- *  - Profile directory: app-<program-name>
+ *  - Profile directory: <program-name>
  *  - Desktop file: <profile-dir>/<program-name>.desktop
  *
  * System web applications have a profile dir without a desktop file.
@@ -89,7 +91,7 @@ get_app_profile_directory_name (const char *id)
   char *profile_dir;
   char *encoded;
 
-  profile_dir = g_strconcat (EPHY_WEB_APP_PREFIX, EPHY_WEB_APP_PROGRAM_NAME_PREFIX, id, NULL);
+  profile_dir = g_strconcat (EPHY_WEB_APP_PROGRAM_NAME_PREFIX, id, NULL);
   encoded = get_encoded_path (profile_dir);
   g_free (profile_dir);
 
@@ -109,18 +111,30 @@ get_app_desktop_filename (const char *id)
   return encoded;
 }
 
-static const char *
-get_program_name_from_profile_directory (const char *profile_dir)
+const char *
+ephy_web_application_get_program_name_from_profile_directory (const char *profile_dir)
 {
   const char *name;
 
-  name = strstr (profile_dir, EPHY_WEB_APP_PREFIX);
-  if (!name) {
-    g_warning ("Profile directory %s does not begin with required web app prefix %s", profile_dir, EPHY_WEB_APP_PREFIX);
+  // Just get the basename
+  name = strrchr (profile_dir, G_DIR_SEPARATOR);
+  if (name == NULL) {
+    g_warning ("Profile directoroy %s is not a valid path", profile_dir);
     return NULL;
   }
 
-  return name + strlen (EPHY_WEB_APP_PREFIX);
+  name++; // Strip '/'
+
+  // Legacy web app support
+  if (g_str_has_prefix (name, "app-"))
+    name += strlen ("app-");
+
+  if (!g_str_has_prefix (name, EPHY_WEB_APP_PROGRAM_NAME_PREFIX)) {
+    g_warning ("Profile directory %s does not begin with required web app prefix %s", profile_dir, EPHY_WEB_APP_PROGRAM_NAME_PREFIX);
+    return NULL;
+  }
+
+  return name;
 }
 
 static const char *
@@ -139,10 +153,23 @@ get_app_id_from_profile_directory (const char *profile_dir)
 {
   const char *program_name;
 
-  program_name = get_program_name_from_profile_directory (profile_dir);
+  program_name = ephy_web_application_get_program_name_from_profile_directory (profile_dir);
   return program_name ? get_app_id_from_program_name (program_name) : NULL;
 }
 
+static char *
+ephy_web_application_get_directory_under (const char *id,
+                                          const char *path)
+{
+  g_autofree char *app_dir = NULL;
+
+  app_dir = get_app_profile_directory_name (id);
+  if (!app_dir)
+    return NULL;
+
+  return g_build_filename (path, app_dir, NULL);
+}
+
 /**
  * ephy_web_application_get_profile_directory:
  * @id: the application identifier
@@ -154,18 +181,19 @@ get_app_id_from_profile_directory (const char *profile_dir)
 char *
 ephy_web_application_get_profile_directory (const char *id)
 {
-  char *dot_dir, *app_dir, *profile_dir;
+  return ephy_web_application_get_directory_under (id, g_get_user_data_dir ());
+}
 
-  app_dir = get_app_profile_directory_name (id);
-  if (!app_dir)
-    return NULL;
+static char *
+ephy_web_application_get_cache_directory (const char *id)
+{
+  return ephy_web_application_get_directory_under (id, g_get_user_cache_dir ());
+}
 
-  dot_dir = !ephy_dot_dir_is_default () ? ephy_default_dot_dir () : NULL;
-  profile_dir = g_build_filename (dot_dir ? dot_dir : ephy_dot_dir (), app_dir, NULL);
-  g_free (app_dir);
-  g_free (dot_dir);
-
-  return profile_dir;
+static char *
+ephy_web_application_get_config_directory (const char *id)
+{
+  return ephy_web_application_get_directory_under (id, g_get_user_config_dir ());
 }
 
 /**
@@ -180,51 +208,68 @@ ephy_web_application_get_profile_directory (const char *id)
 gboolean
 ephy_web_application_delete (const char *id)
 {
-  char *profile_dir;
-  char *desktop_file = NULL, *desktop_path = NULL;
-  GFile *launcher = NULL;
-  gboolean return_value = FALSE;
+  g_autofree char *profile_dir = NULL;
+  g_autofree char *cache_dir = NULL;
+  g_autofree char *config_dir = NULL;
+  g_autofree char *desktop_file = NULL;
+  g_autofree char *desktop_path = NULL;
+  g_autoptr(GFile) launcher = NULL;
+  g_autoptr(GError) error = NULL;
 
   g_assert (id);
 
   profile_dir = ephy_web_application_get_profile_directory (id);
   if (!profile_dir)
-    goto out;
+    return FALSE;
 
   /* If there's no profile dir for this app, it means it does not
    * exist. */
   if (!g_file_test (profile_dir, G_FILE_TEST_IS_DIR)) {
     g_warning ("No application with id '%s' is installed.\n", id);
-    goto out;
+    return FALSE;
   }
 
-  if (!ephy_file_delete_dir_recursively (profile_dir, NULL))
-    goto out;
+  if (!ephy_file_delete_dir_recursively (profile_dir, &error)) {
+    g_warning ("Failed to recursively delete %s: %s", profile_dir, error->message);
+    return FALSE;
+  }
   LOG ("Deleted application profile.\n");
 
+  cache_dir = ephy_web_application_get_cache_directory (id);
+  if (g_file_test (cache_dir, G_FILE_TEST_IS_DIR)) {
+    if (!ephy_file_delete_dir_recursively (cache_dir, &error)) {
+      g_warning ("Failed to recursively delete %s: %s", cache_dir, error->message);
+      return FALSE;
+    }
+    LOG ("Deleted application cache directory.\n");
+  }
+
+  config_dir = ephy_web_application_get_config_directory (id);
+  if (g_file_test (config_dir, G_FILE_TEST_IS_DIR)) {
+    if (!ephy_file_delete_dir_recursively (config_dir, &error)) {
+      g_warning ("Failed to recursively delete %s: %s", config_dir, error->message);
+      return FALSE;
+    }
+    LOG ("Deleted application config directory.\n");
+  }
+
   desktop_file = get_app_desktop_filename (id);
-  if (!desktop_file)
-    goto out;
+  if (!desktop_file) {
+    g_warning ("Failed to compute desktop filename for app %s", id);
+    return FALSE;
+  }
 
   desktop_path = g_build_filename (g_get_user_data_dir (), "applications", desktop_file, NULL);
   if (g_file_test (desktop_path, G_FILE_TEST_IS_SYMLINK)) {
     launcher = g_file_new_for_path (desktop_path);
-    if (!g_file_delete (launcher, NULL, NULL))
-      goto out;
+    if (!g_file_delete (launcher, NULL, &error)) {
+      g_warning ("Failed to delete %s: %s", desktop_path, error->message);
+      return FALSE;
+    }
     LOG ("Deleted application launcher.\n");
   }
-  return_value = TRUE;
 
- out:
-
-  g_free (profile_dir);
-
-  if (launcher)
-    g_object_unref (launcher);
-  g_free (desktop_file);
-  g_free (desktop_path);
-
-  return return_value;
+  return TRUE;
 }
 
 static char *
@@ -329,33 +374,42 @@ ephy_web_application_create (const char *id,
                              const char *name,
                              GdkPixbuf  *icon)
 {
-  char *profile_dir;
-  char *desktop_file_path = NULL;
+  g_autofree char *app_file = NULL;
+  g_autofree char *profile_dir = NULL;
+  g_autofree char *desktop_file_path = NULL;
 
   /* If there's already a WebApp profile for the contents of this
    * view, do nothing. */
   profile_dir = ephy_web_application_get_profile_directory (id);
   if (g_file_test (profile_dir, G_FILE_TEST_IS_DIR)) {
-    LOG ("Profile directory %s already exists", profile_dir);
-    goto out;
+    g_warning ("Profile directory %s already exists", profile_dir);
+    return NULL;
   }
 
   /* Create the profile directory, populate it. */
-  if (g_mkdir (profile_dir, 488) == -1) {
-    LOG ("Failed to create directory %s", profile_dir);
-    goto out;
+  if (g_mkdir_with_parents (profile_dir, 488) == -1) {
+    g_warning ("Failed to create directory %s", profile_dir);
+    return NULL;
   }
 
+  /* Skip migration for new web apps. */
+  ephy_profile_utils_set_migration_version_for_profile_dir (EPHY_PROFILE_MIGRATION_VERSION, profile_dir);
+
+  /* Create an .app file. */
+  app_file = g_build_filename (profile_dir, ".app", NULL);
+  int fd = g_open (app_file, O_WRONLY|O_CREAT|O_TRUNC, 0644);
+  if (fd < 0) {
+    g_warning ("Failed to create .app file: %s", g_strerror (errno));
+    return NULL;
+  }
+  close (fd);
+
   /* Create the deskop file. */
   desktop_file_path = create_desktop_file (id, name, address, profile_dir, icon);
   if (desktop_file_path)
     ephy_web_application_initialize_settings (profile_dir);
 
- out:
-  if (profile_dir)
-    g_free (profile_dir);
-
-  return desktop_file_path;
+  return g_steal_pointer (&desktop_file_path);
 }
 
 char *
@@ -391,7 +445,7 @@ ephy_web_application_setup_from_profile_directory (const char *profile_directory
 
   g_assert (profile_directory != NULL);
 
-  program_name = get_program_name_from_profile_directory (profile_directory);
+  program_name = ephy_web_application_get_program_name_from_profile_directory (profile_directory);
   if (!program_name)
     exit (1);
 
@@ -534,6 +588,63 @@ ephy_web_application_for_profile_directory (const char *profile_dir)
   return app;
 }
 
+static GList *
+ephy_web_application_get_application_list_internal (gboolean only_legacy)
+{
+  GFileEnumerator *children = NULL;
+  GFileInfo *info;
+  GList *applications = NULL;
+  g_autofree char *parent_directory_path = NULL;
+  g_autoptr(GFile) parent_directory = NULL;
+
+  if (only_legacy)
+    parent_directory_path = g_build_filename (g_get_user_config_dir (), "epiphany", NULL);
+  else
+    parent_directory_path = g_strdup (g_get_user_data_dir ());
+
+  parent_directory = g_file_new_for_path (parent_directory_path);
+  children = g_file_enumerate_children (parent_directory,
+                                        "standard::name",
+                                        0, NULL, NULL);
+  if (!children)
+    return NULL;
+
+  info = g_file_enumerator_next_file (children, NULL, NULL);
+  while (info) {
+    const char *name;
+
+    name = g_file_info_get_name (info);
+    if ((only_legacy && g_str_has_prefix (name, "app-")) ||
+        (!only_legacy && g_str_has_prefix (name, EPHY_WEB_APP_PROGRAM_NAME_PREFIX))) {
+      EphyWebApplication *app;
+      char *profile_dir;
+
+      profile_dir = g_build_filename (parent_directory_path, name, NULL);
+      app = ephy_web_application_for_profile_directory (profile_dir);
+      if (app) {
+        if (!only_legacy) {
+          g_autofree char *app_file = g_build_filename (profile_dir, ".app", NULL);
+          if (g_file_test (app_file, G_FILE_TEST_EXISTS))
+            applications = g_list_prepend (applications, app);
+          else
+            g_object_unref (app);
+        } else
+          applications = g_list_prepend (applications, app);
+      }
+
+      g_free (profile_dir);
+    }
+
+    g_object_unref (info);
+
+    info = g_file_enumerator_next_file (children, NULL, NULL);
+  }
+
+  g_object_unref (children);
+
+  return g_list_reverse (applications);
+}
+
 /**
  * ephy_web_application_get_application_list:
  *
@@ -546,46 +657,28 @@ ephy_web_application_for_profile_directory (const char *profile_dir)
 GList *
 ephy_web_application_get_application_list (void)
 {
-  GFileEnumerator *children = NULL;
-  GFileInfo *info;
-  GList *applications = NULL;
-  GFile *dot_dir;
-  char *default_dot_dir;
-
-  default_dot_dir = !ephy_dot_dir_is_default () ? ephy_default_dot_dir () : NULL;
-  dot_dir = g_file_new_for_path (default_dot_dir ? default_dot_dir : ephy_dot_dir ());
-  children = g_file_enumerate_children (dot_dir,
-                                        "standard::name",
-                                        0, NULL, NULL);
-  g_object_unref (dot_dir);
-
-  info = g_file_enumerator_next_file (children, NULL, NULL);
-  while (info) {
-    const char *name;
-
-    name = g_file_info_get_name (info);
-    if (g_str_has_prefix (name, EPHY_WEB_APP_PREFIX)) {
-      EphyWebApplication *app;
-      char *profile_dir;
-
-      profile_dir = g_build_filename (default_dot_dir ? default_dot_dir : ephy_dot_dir (), name, NULL);
-      app = ephy_web_application_for_profile_directory (profile_dir);
-      if (app)
-        applications = g_list_prepend (applications, app);
-      g_free (profile_dir);
-    }
-
-    g_object_unref (info);
-
-    info = g_file_enumerator_next_file (children, NULL, NULL);
-  }
-
-  g_object_unref (children);
-  g_free (default_dot_dir);
-
-  return g_list_reverse (applications);
+  return ephy_web_application_get_application_list_internal (FALSE);
 }
 
+/**
+ * ephy_web_application_get_legacy_application_list:
+ *
+ * Gets a list of the currently installed web applications.
+ * This is only used for the profile migrator as it gets
+ * applications in the legacy directory.
+ *
+ * Free the returned GList with
+ * ephy_web_application_free_application_list.
+ *
+ * Returns: (transfer-full): a #GList of #EphyWebApplication objects
+ **/
+GList *
+ephy_web_application_get_legacy_application_list (void)
+{
+  return ephy_web_application_get_application_list_internal (TRUE);
+}
+
+
 /**
  * ephy_web_application_free_application_list:
  * @list: an #EphyWebApplication GList
@@ -662,14 +755,43 @@ ephy_web_application_initialize_settings (const char *profile_directory)
   g_free (name);
 }
 
-gboolean
-ephy_web_application_is_uri_allowed (const char* uri)
+static gboolean
+urls_have_same_origin (const char *a_url,
+                       const char *b_url)
 {
+  SoupURI *a_uri, *b_uri;
+  gboolean retval = FALSE;
+
+  a_uri = soup_uri_new (a_url);
+  if (!a_uri)
+    return retval;
+
+  b_uri = soup_uri_new (b_url);
+  if (b_uri) {
+    retval = a_uri->host && b_uri->host && soup_uri_host_equal (a_uri, b_uri);
+    soup_uri_free (b_uri);
+  }
+
+  soup_uri_free (a_uri);
+
+  return retval;
+}
+
+gboolean
+ephy_web_application_is_uri_allowed (const char *uri)
+{
+  EphyWebApplication *webapp = ephy_web_application_for_profile_directory (ephy_profile_dir ());
   SoupURI *request_uri;
   char **urls;
   guint i;
   gboolean matched = FALSE;
 
+  if (g_str_has_prefix (uri, "blob:") || g_str_has_prefix (uri, "data:"))
+    return TRUE;
+
+  if (urls_have_same_origin (uri, webapp->url))
+    return TRUE;
+
   if (g_strcmp0 (uri, "about:blank") == 0)
     return TRUE;
 
